import numpy as np


def get_comparable_pairs_npy(time, event):
    y = time
    e = event

    grid_x, grid_y = np.meshgrid(y, y)
    grid_x = np.tril(grid_x)
    grid_y = np.tril(grid_y)
    diff_truth = grid_x - grid_y

    grid_ex, grid_ey = np.meshgrid(e, e)
    valid_pairs = ((diff_truth < 0) & (grid_ex == 1))
    res1 = np.stack(np.where(valid_pairs)).T

    diff_truth = grid_y - grid_x
    valid_pairs = ((diff_truth < 0) & (grid_ey == 1))
    res2 = np.flip(np.stack(np.where(valid_pairs)).T, axis=1)

    pairs = np.concatenate([res1, res2])
    return pairs


def cindex_by_pair_npy(v_high, v_low):
    eval_comparable = (v_high < v_low).astype('float32')
    eval_non_comparable = (v_high == v_low).astype('float32')
    return eval_comparable + (eval_non_comparable * 0.5)


def compute_univariate_score_of_npy(pairs, biom_values):
    biom_values_low = biom_values[pairs[:, 0]]
    biom_values_high = biom_values[pairs[:, 1]]

    # compute feature's signs
    features_cindex_by_pair = cindex_by_pair_npy(biom_values_high, biom_values_low)
    features_cindex = 1. - features_cindex_by_pair.mean(axis=0)

    return features_cindex


def compute_univariate_score_npy(df, candidates, target_columns, biom_values):
    candidates_id = np.arange(len(candidates))

    max_chunk_size = 1e4
    if len(candidates_id) > max_chunk_size:
        n_chunks = len(candidates_id) / max_chunk_size
        chunks = np.array_split(candidates_id, n_chunks)
    else:
        chunks = [candidates_id]

    all_scores_by_target = {}
    for target in target_columns:
        # get comparable train pairs values
        pairs = get_comparable_pairs_npy(df[target_columns[target][0]].values,
                                         df[target_columns[target][1]].astype('bool').values)
        univ_scores = []
        for chunk in chunks:
            univ_scores.append(compute_univariate_score_of_npy(pairs, biom_values[:, chunk]))
        all_scores_by_target[target] = np.concatenate(univ_scores)

    return all_scores_by_target


def compute_pvals_target_npy(scores, random_scores):
    scores = abs(scores - 0.5)
    n_random = len(random_scores)

    max_chunk_size = 1e2

    if len(scores) > max_chunk_size:
        n_chunks = int(len(scores) / max_chunk_size)
        chunks = np.array_split(scores, n_chunks)
    else:
        chunks = [scores]

    all_pvals = []
    for chunk_values in chunks:
        chunk_values = chunk_values[:, None]
        pvals = ((random_scores >= chunk_values).sum(axis=1) + 1) / (n_random + 1)
        all_pvals.append(pvals)
    all_pvals = np.concatenate(all_pvals)
    return all_pvals


def compute_pvals_npy(scores, random_scores):
    pvals = {}
    for target in scores:
        pvals[target] = compute_pvals_target_npy(scores[target], random_scores[target])
    return pvals



from matplotlib import pyplot as plt
from multipy.fdr import tst
from tqdm import tqdm
import seaborn as sns
from robi.evaluation import univariate_evaluation, score_of_random
from robi.preselection import primary_selection
from robi.utils import *
from joblib import Parallel, delayed

sns.set_theme(style="whitegrid")

import time
import robi
import torch


def get_best_candidates_per_cluster(scores, clusters):
    representative = []
    for cluster in clusters:
        if len(cluster) > 1:
            best_idx = np.argmin(scores[cluster])
            representative.append(cluster[best_idx])
        else:
            representative.append(cluster[0])
    return np.array(representative)


def get_sel_by_permissiveness(pvals, corr_clusters, targets):
    all_n_sel = []

    for target in targets:
        pvals_target = pvals[target]

        candidates = get_best_candidates_per_cluster(pvals_target, corr_clusters)
        pvals_target = pvals_target[candidates]

        for permissiveness in np.arange(0.01, 1.01, 0.01).round(2):
            selected = candidates[tst(pvals_target, q=permissiveness)]
            all_n_sel.append({
                'permissiveness': permissiveness,
                'target': target,
                'n_selected': len(selected),
                'selected': selected,
            })
    return pd.DataFrame(all_n_sel).set_index(['target', 'permissiveness'])


def get_nfp_by_permissiveness(pvals_permut, corr_clusters, targets, n_jobs):
    res = Parallel(n_jobs=n_jobs)(delayed(get_sel_by_permissiveness)
                                  (s, corr_clusters, targets)
                                  for s in pvals_permut)
    res = pd.concat(res)
    res = res.reset_index()
    res = res.drop(columns=['selected'])
    res = res.rename(columns={'n_selected': 'n_FP'})
    res = res.set_index(['target', 'permissiveness'])
    return res


def get_permissiveness_effect(pval, pvals_permut, corr_clusters, targets, n_workers):
    list_targets = list(targets.keys())
    sel_by_permissiveness = get_sel_by_permissiveness(pval, corr_clusters, list_targets)
    nfp_by_permissiveness = get_nfp_by_permissiveness(pvals_permut, corr_clusters, list_targets, n_workers)
    return sel_by_permissiveness, nfp_by_permissiveness


def plot_permissiveness_effect(sel_by_permissiveness, nfp_by_permissiveness, targets):
    for target in targets:
        sns.lineplot(sel_by_permissiveness.loc[target], x='permissiveness', y='n_selected',
                     label='Number of\nselected candidates', color='#ec6602')
        sns.lineplot(nfp_by_permissiveness.loc[target], x='permissiveness', y='n_FP',
                     label='Number of\nfalse positives', errorbar=('pi', 95), color='#009999')
        plt.title(target)
        plt.ylabel('')
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        plt.show()


def make_selection(df,
                   candidates,
                   targets,
                   confounders=None,
                   known=None,
                   strata=None,
                   max_corr_cluster=1.,
                   n_uni_pval=1000,
                   n_fp_estimate=1000,
                   verbose=True,
                   n_jobs=1,
                   device="cpu"):
    if verbose:
        print('Selection started...')

    use_torch = torch_installed(verbose=True)
    if use_torch and device != 'cuda':
        print(f"device is {device}. If you have a GPU, switching to cuda will provide significant speed gains.")

    if confounders is None:
        confounders = []
    if known is None:
        known = []
    if isinstance(strata, list):
        if len(strata) == 0:
            strata = None
    confounders = list(set(confounders + known))

    if isinstance(candidates, np.ndarray):
        candidates = candidates.tolist()

    df, targets = format_targets(df, targets)

    # drop cases where confounders are not defined
    if len(confounders) > 0:
        df = df[df[confounders].isna().sum(axis=1) == 0]

    if strata is None:
        to_norm = candidates + confounders
    else:
        to_norm = list(set(candidates + confounders) - set(strata))
    df = normalize_columns(df, to_norm)

    if verbose and len(confounders) > 0:
        check_confounders(df, confounders, targets, strata)

    # perform the preselection
    candidates, corr_clusters, candidates_correlations = primary_selection(df,
                                                                           candidates,
                                                                           confounders,
                                                                           strata,
                                                                           known,
                                                                           targets,
                                                                           max_corr_cluster,
                                                                           verbose,
                                                                           n_jobs)

    scores_random = score_of_random(df, targets, device, n_uni_pval, use_torch)
    scores, pvals = univariate_evaluation(df, candidates, targets, device, scores_random, use_torch)

    if verbose:
        check_n_random(pvals, n_uni_pval, use_torch)

    pvals_permut = robi.evaluation.get_permuted_scores(df, candidates, scores_random, targets, device, n_fp_estimate,
                                                       verbose, use_torch)

    sel_by_permissiveness, nfp_by_permissiveness = get_permissiveness_effect(pvals, pvals_permut,
                                                                             corr_clusters, targets,
                                                                             n_jobs)
    if verbose:
        plot_permissiveness_effect(sel_by_permissiveness, nfp_by_permissiveness, targets)

    nfp_by_permissiveness['n_selected'] = sel_by_permissiveness.loc[nfp_by_permissiveness.index, 'n_selected']
    nfp_by_permissiveness['P_only_FP'] = nfp_by_permissiveness['n_FP'] >= nfp_by_permissiveness['n_selected']

    nfp_by_permissiveness_g = nfp_by_permissiveness.groupby(['target', 'permissiveness'])
    nfp_mean = nfp_by_permissiveness_g['n_FP'].mean().round(1)
    nfp_cil = nfp_by_permissiveness_g['n_FP'].quantile(0.025).round(2)
    nfp_cih = nfp_by_permissiveness_g['n_FP'].quantile(0.975).round(2)
    nfp_mean = nfp_mean.round(1).astype('str') + ' (' + nfp_cil.astype('str') + '-' + nfp_cih.astype('str') + ')'
    p_only_FP = nfp_by_permissiveness_g['P_only_FP'].mean()

    sel_by_permissiveness = pd.concat([sel_by_permissiveness, nfp_mean, p_only_FP], axis=1)

    sel_by_permissiveness = sel_by_permissiveness[['n_selected', 'n_FP', 'P_only_FP', 'selected']]

    scores_col = []
    pval_col = []
    target_col = []
    cand_col = []
    for target in scores:
        s = scores[target]
        p = pvals[target]
        if use_torch:
            s = s.cpu().numpy()
            p = p.cpu().numpy()
        s = s.tolist()
        p = p.tolist()
        scores_col += s
        pval_col += p
        target_col += [target] * len(s)
        cand_col += candidates

    scores = pd.DataFrame(data={
        'candidate': cand_col,
        'target': target_col,
        'C_index': scores_col,
        'p_value': pval_col,
    })
    scores = scores.set_index(['candidate', 'target'])

    candidates = np.array(candidates)
    sel_by_permissiveness['selected'] = [candidates[x] for x in sel_by_permissiveness['selected']]

    torch.cuda.empty_cache()
    return sel_by_permissiveness, scores



from lifelines import CoxPHFitter
from lifelines.statistics import proportional_hazard_test
import scipy.optimize as opt
from sklearn.datasets import make_regression
import numpy as np
import pandas as pd


def check_n_random(pvals, n_random, use_torch):
    """
    Check if the number of random features is sufficient based on the minimum p-value.
    The number of random features that are better than the candidates should be high enough
    to ensure precise p-value estimation, event for candidates with really high score.
    Prints a warning if the number of random features is considered insufficient.
    """

    # Calculate the minimum p-value across all
    if use_torch:
        min_pval = np.min([pvals[t].min().numpy()[()] for t in pvals])
    else:
        min_pval = np.min([np.min(pvals[t]) for t in pvals])

    n_sup = int(min_pval * n_random)

    # Check if the number of superior random features is less than 200
    if n_sup < 200:
        if n_sup > 1:
            recommended_n_random = int(200 / n_sup * n_random)
            print(f"Warning: only {n_sup} random features were better than the best candidate."
                  f"\nConsider increasing n_random. We recommend at least {recommended_n_random:.2E} random features.")
        else:
            print('Warning: no random features were better than the best candidate. Consider increasing n_random.')
    else:
        print('n_random is high enough')


def normalize_columns(df, columns):
    """
    Normalize (z-score) specified columns in the dataframe.
    """

    # # Check if the specified columns exist in the dataframe
    # if not all(col in df.columns for col in columns):
    #     raise ValueError("One or more specified columns do not exist in the dataframe.")
    #
    # # Create a new dataframe with the normalized columns
    normalized_df = df.copy()
    #
    # # Calculate the z-score for each column
    # std = df[columns].std()
    # std[(std == 0) | (np.isnan(std))] = 1
    normalized_df.loc[:, columns] = (df[columns] - df[columns].mean()) / df[columns].std()

    return normalized_df


def check_proportional_hazard_for_target(df, target, targets, confounders, strata):
    """
    Check proportional hazard assumption for a specific target.
    """

    # Construct column names based on target
    duration_col = targets[target][0]
    event_col = targets[target][1]

    # Filter necessary columns from dataframe
    df_filtered = df[confounders + [duration_col, event_col]]

    # Fit the Cox Proportional Hazard model
    cph = CoxPHFitter()
    cph.fit(df_filtered, duration_col=duration_col, event_col=event_col, strata=strata, robust=True)

    # Test for proportional hazard assumption
    results = proportional_hazard_test(cph, df_filtered)

    # Check if any confounder violates the assumption
    for confounder, row in results.summary.iterrows():
        if row['p'] < 0.05:
            print(f"{confounder} violates proportional hazard assumptions for {target} (p<{row['p']:.2f}).",
                  f"Consider adding {confounder} to the strata list.")
            return False

    return True


def check_proportional_hazard(df, confounders, targets, strata):
    """
    Verify the proportional hazard assumption for specified targets and confounders.
    Prints the results of the verification.
    """

    all_tests_ok = all(
        check_proportional_hazard_for_target(df, target, targets, confounders, strata) for target in targets)

    if all_tests_ok:
        print('Confounders follow proportional hazard assumptions')


def check_confounders(df, confounders, targets, strata):
    """
    Check if the number of confounders exceeds the allowed limit for each target and
    verify if the confounders follows the proportional hazard assumption.
    """

    print('\nVerifying confounders settings...')

    # Calculate the maximum allowed number of confounders for each target
    # Use the rule-of-thumb stating that you should have at least
    for target, (_, event_col) in targets.items():
        n_event = df[event_col].sum()
        max_n_conf = int(n_event / 8 + 0.5)

        if len(confounders) > max_n_conf:
            print(f'Too many confounders for {target}. Max recommended: {max_n_conf}')

    # Check for proportional hazards (function not provided in the original code)
    check_proportional_hazard(df, confounders, targets, strata)


def torch_installed(verbose=False):
    try:
        import torch
        return True
    except:
        if verbose:
            print(
                'Could not load PyTorch. Consider installing or fixing it for substantial speed gains. Using Numpy instead. ')
        return False


def format_targets(df, targets):
    """
    :param df:
    :param targets: possible formats
        't1'
        ['t1']
        ['t1', 't2]
        {
            't1': ('time', 'event'),
        }
        {
            't1': ('time1', 'event1'),
            't2': ('time2', 'event2'),
        }
        {
            't1': ('time1'),
            't2': ('time2', 'event2'),
        }
        {
            't1': 'time1',
            't2': ('time2', 'event2'),
        }
    :return:
    """
    df = df.copy()
    invalid = False

    if isinstance(targets, str):
        df['events'] = True
        targets = {
            targets: (targets, 'events')
        }

    if isinstance(targets, list):
        new_targets = {}
        for t in targets:
            df[t + '_events'] = True
            new_targets[t] = (t, t + '_events')
        targets = new_targets

    if isinstance(targets, dict):
        for t in targets:
            if isinstance(targets[t], str):
                df[t + '_events'] = True
                targets[t] = (targets[t], t + '_events')
            elif isinstance(targets[t], tuple) or isinstance(targets[t], list):
                if len(targets[t]) != 2:
                    if len(targets[t]) == 1:
                        df[t + '_events'] = True
                        targets[t] = (targets[t], t + '_events')
                    else:
                        invalid = True
            else:
                invalid = True

    if invalid:
        raise ValueError('Invalid target format. See the README for examples.')

    return df, targets


def new_synthetic_dataset(n_samples, censoring, nb_features, n_informative, effective_rank, noise):
    X, Y, coef = make_regression(n_samples=n_samples,
                                 n_features=nb_features,
                                 n_informative=n_informative,
                                 effective_rank=effective_rank,
                                 noise=noise,
                                 coef=True)

    Y += abs(np.min(Y))

    def get_observed_time(x):
        rnd_cens = np.random.RandomState(0)
        # draw censoring times
        time_censor = rnd_cens.uniform(high=x, size=n_samples)
        event = Y < time_censor
        time = np.where(event, Y, time_censor)
        return event, time

    def censoring_amount(x):
        event, _ = get_observed_time(x)
        cens = 1.0 - event.sum() / event.shape[0]
        return (cens - censoring) ** 2

    # search for upper limit to obtain the desired censoring amount
    res = opt.minimize_scalar(censoring_amount,
                              method="bounded",
                              bounds=(0, Y.max()))

    # compute observed time
    event, time = get_observed_time(res.x)

    # upper time limit such that the probability
    # of being censored is non-zero for `t > tau`
    tau = time[event].max()
    mask = time < tau
    X = X[mask]
    Y = Y[mask]
    event = event[mask]

    df = pd.DataFrame(data=X)
    df['time'] = Y
    df['event'] = event
    df.columns = df.columns.values.astype(str)

    return df, coef



import pandas as pd
import robi

def test_dlbcl():
    df = pd.read_csv('../data/DLBCL_test.csv')

    targets = {
      'PFS': ('PFS_months', 'PFS_event_happened'),
      'OS': ('OS_months', 'OS_event_happened'),
    }
    confounders = ['ECOG', 'aaIPI', 'treatment', 'sex']
    strata = ['treatment']

    res, _ = robi.make_selection(df,
                    candidates=['TMTV', 'Dmax'],
                    targets = targets,
                    confounders=confounders,
                    strata =strata,
                    verbose=False)
    assert res[res['n_sel_PFS']>0].iloc[0]['n_sel_PFS'] == 2


    res, _ = robi.make_selection(df,
                    candidates=['TMTV', 'Dmax'],
                    targets = targets,
                    confounders=['TMTV']+confounders,
                    strata =strata,
                    verbose=False)
    assert res[res['n_sel_PFS']>0].iloc[0]['n_sel_PFS'] == 1


from robi.evaluation_npy import *
from robi.evaluation_torch import *
from tqdm import tqdm
import pandas as pd


def compute_pvals(scores, scores_random, device, use_torch):
    if use_torch:
        return compute_pvals_torch(scores, scores_random, device)
    return compute_pvals_npy(scores, scores_random)


def compute_univariate_score(df, random_cols, targets, device, use_torch):
    if use_torch:
        df_arr = torch.as_tensor(df[random_cols].values, device=device, dtype=torch.float)
        return compute_univariate_score_torch(df, random_cols, targets, df_arr)
    df_arr = df[random_cols].values
    return compute_univariate_score_npy(df, random_cols, targets, df_arr)


def score_of_random(df, targets, device, n_random, use_torch):
    df_random = pd.DataFrame(data=np.random.uniform(0, 1, (df.shape[0], int(n_random))), index=df.index)
    random_cols = df_random.columns
    df_random = pd.concat([df[sum([list(targets[x]) for x in targets], [])], df_random], axis=1)
    random_scores = compute_univariate_score(df_random, random_cols, targets, device, use_torch)
    for target in random_scores:
        random_scores[target] = abs(random_scores[target] - 0.5)
    return random_scores


def univariate_evaluation(df, candidates, targets, device, scores_random, use_torch):
    scores = compute_univariate_score(df, candidates, targets, device, use_torch)
    pvals = compute_pvals(scores, scores_random, device, use_torch)
    return scores, pvals


def get_permuted_scores(df, candidates, scores_random_by_target, targets, device, n_permut, verbose, use_torch):
    cols_to_permut = sum([list(targets[x]) for x in targets], [])
    all_pscores = []

    if use_torch:
        df_arr = torch.as_tensor(df[candidates].values, device=device, dtype=torch.float)
    else:
        df_arr = df[candidates].values

    for _ in tqdm(range(n_permut),
                  desc='Computing scores of permutations',
                  bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',
                  disable=not verbose):
        dfp = df.copy()
        dfp.loc[dfp.index, cols_to_permut] = dfp.loc[np.random.permutation(dfp.index), cols_to_permut].values

        if use_torch:
            scores = compute_univariate_score_torch(dfp, candidates, targets, df_arr)
        else:
            scores = compute_univariate_score_npy(dfp, candidates, targets, df_arr)

        pscores = compute_pvals(scores, scores_random_by_target, device, use_torch)
        all_pscores.append(pscores)
    return all_pscores



import pandas as pd
import numpy as np
from lifelines.utils import concordance_index
from robi.evaluation_torch import compute_univariate_score_torch, compute_pvals_torch
from robi.evaluation_npy import compute_univariate_score_npy, compute_pvals_npy
from robi.evaluation import score_of_random

def test_evaluation():
    df = pd.read_csv('../data/DLBCL_test.csv')
    targets = {
      'PFS': ('PFS_months', 'PFS_event_happened'),
    }

    n_fakes = 100

    df_random = pd.DataFrame(data=np.random.uniform(0,1, (df.shape[0], int(n_fakes))), index=df.index)
    df_random.columns = df_random.columns.astype('str')
    random_cols = df_random.columns
    df = pd.concat([df, df_random], axis=1)

    ref_cindex = []
    for c in random_cols:
        ref_cindex.append(concordance_index(event_times=df['PFS_months'],
                                       predicted_scores=-df[c],
                                       event_observed=df['PFS_event_happened']))
    ref_cindex = np.array(ref_cindex)

    scores_npy = compute_univariate_score_npy(df, random_cols, targets)
    assert np.allclose(ref_cindex, scores_npy['C_index'].values, atol=1e-3)

    scores_torch = compute_univariate_score_torch(df, random_cols, targets, device='cuda')
    assert np.allclose(ref_cindex, scores_torch['C_index'].values, atol=1e-3)

    scores_random = score_of_random(df, targets, 'cuda', 1000)
    pval_npy = compute_pvals_npy(scores_npy, scores_random)
    pval_torch = compute_pvals_torch(scores_npy, scores_random, 'cuda')

    assert np.allclose(pval_npy['p_value'].values, pval_torch['p_value'].values, atol=1e-3)
